from lona.html.abstract_node import AbstractNode
from lona.protocol import NODE_TYPE


class TextNode(AbstractNode):
    def __init__(self, string):
        self._string = str(string)

    def wrap_method(self, method):
        def wrapper(*args, **kwargs):
            return_value = method(*args, **kwargs)

            if isinstance(return_value, str):
                return TextNode(return_value)

            return return_value

        return wrapper

    def __getattribute__(self, name):
        if name in ('id', '_string'):
            return super().__getattribute__(name)

        if hasattr(self, '_string') and hasattr(self._string, name):
            attribute = getattr(self._string, name)

            if callable(attribute):
                return self.wrap_method(attribute)

        return super().__getattribute__(name)

    def __dir__(self):
        return dir(self._string)

    def __add__(self, other):
        string = other

        if isinstance(string, TextNode):
            string = string._string

        return TextNode(self._string + string)

    def __getitem__(self, key):
        return TextNode(self._string[key])

    def __iadd__(self, other):
        raise TypeError('unsupported operation: +=')

    def __str__(self):
        return self._string

    def __len__(self):
        return len(self._string)

    def __bool__(self):
        return True

    def __repr__(self):
        return f'<TextNode({self._string!r})>'

    # serialization ###########################################################
    def _serialize(self):
        return [NODE_TYPE.TEXT_NODE, self.id, self._string]

    # node helper #############################################################
    def remove(self):
        if not self._parent:
            raise RuntimeError('node has no parent node')

        self._parent.remove(self)

    def get_text(self):
        return self._string



import logging
import asyncio
import socket
import signal
import os

from aiohttp.web import run_app
import aiohttp

from lona.shell.shell import generate_shell_server, embed_shell
from lona.worker_pool import WorkerPool
from lona.logging import setup_logging
from lona.server import LonaServer

AIOHTTP_VERSION = tuple(
    int(part) for part in aiohttp.__version__.split('.')[:2]
)

logger = logging.getLogger('lona')


def run_server(args, server=None):
    loop = asyncio.get_event_loop()

    # setup logging
    log_formatter, log_filter = setup_logging(args)

    # setup lona server
    server = server or LonaServer(
        project_root=args.project_root,
        settings_paths=args.settings,
        settings_pre_overrides=args.settings_pre_overrides,
        settings_post_overrides=args.settings_post_overrides,
    )
    app = server._app

    worker_pool = WorkerPool(
        settings=server.settings,
    )

    async def shutdown(app):
        server = app['lona_server']

        await server.loop.run_in_executor(None, server.worker_pool.shutdown)

    app.on_shutdown.append(shutdown)

    server._loop = loop
    server._worker_pool = worker_pool

    # run server
    if args.shell:
        async def start_shell(server):
            def _start_shell():
                embed_kwargs = {
                    'globals': {
                        'loop': loop,
                        'server': server,
                        'cli_args': vars(args),
                        'log_formatter': log_formatter,
                        'log_filter': log_filter,
                    },
                }

                embed_shell(server=server, **embed_kwargs)

                os.kill(os.getpid(), signal.SIGTERM)

            loop.run_in_executor(None, _start_shell)

        loop.create_task(start_shell(server))

    def _run_app():
        keyword_args = {
            'app': app,
            'host': args.host,
            'port': args.port,
            'shutdown_timeout': args.shutdown_timeout,
        }

        # In aiohttp 3.8 the keyword "loop" was added, which is mandatory if
        # you don’t want aiohttp to start a new event loop but use your
        # previously created one
        if AIOHTTP_VERSION >= (3, 8):
            keyword_args['loop'] = loop

        try:
            run_app(**keyword_args)

        except socket.gaierror as exception:
            if exception.errno not in (-2, ):
                raise

            logger.error('socket.gaierror: %s', exception.args[1])

        except OSError as exception:
            if exception.errno not in (13, 98):
                raise

            logger.error('OSError: %s', exception.args[1])

    if args.shell_server_url:
        embed_kwargs = {
            'bind': args.shell_server_url,
            'globals': {
                'loop': loop,
                'server': server,
                'cli_args': vars(args),
                'log_formatter': log_formatter,
                'log_filter': log_filter,
            },
        }

        repl_server = generate_shell_server(server=server, **embed_kwargs)

        with repl_server:
            _run_app()

    else:
        _run_app()



from __future__ import annotations

from typing import TYPE_CHECKING, overload, TypeVar, Union, cast
from collections.abc import Awaitable, Callable
import threading
import asyncio

from typing_extensions import Literal

from lona.view_runtime import VIEW_RUNTIME_STATE, ViewRuntime
from lona.html.abstract_node import AbstractNode
from lona.events.input_event import InputEvent
from lona.static_files import StaticFile
from lona.connection import Connection
from lona.request import Request

# avoid import cycles
if TYPE_CHECKING:  # pragma: no cover
    from lona.events.view_event import ViewEvent
    from lona.server import LonaServer

T = TypeVar('T')
V = TypeVar('V', bound='LonaView')
H = Union[None, AbstractNode, str]


class LonaView:
    STATIC_FILES: list[StaticFile] = []

    def __init__(
            self,
            server: LonaServer,
            view_runtime: ViewRuntime,
            request: Request,
    ) -> None:
        self._server: LonaServer = server
        self._view_runtime: ViewRuntime = view_runtime
        self._request: Request = request

    # properties ##############################################################
    @property
    def server(self) -> LonaServer:
        return self._server

    @property
    def request(self) -> Request:
        return self._request

    # checks ##################################################################
    def _assert_not_main_thread(self) -> None:
        if threading.current_thread() is threading.main_thread():
            raise RuntimeError(
                'this function is not supposed to run in the main thread')

    def _assert_view_is_interactive(self) -> None:
        if not self._request.route.interactive:
            raise RuntimeError(
                'operation is not supported in non-interactive requests')

    def _assert_view_is_running(self) -> None:
        if self._view_runtime.stop_reason:
            raise self._view_runtime.stop_reason

    # asyncio #################################################################
    def _await_sync(self, awaitable: Awaitable[T]) -> T:
        async def await_awaitable() -> T:
            finished, pending = await asyncio.wait(
                [
                    cast(Awaitable, self._view_runtime.stopped),
                    awaitable,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            for finished_future in finished:
                if finished_future is not self._view_runtime.stopped:
                    return finished_future.result()

            if self._view_runtime.stopped in finished:
                for pending_future in pending:
                    if not pending_future.done():
                        pending_future.cancel()

            raise self._view_runtime.stop_reason

        return asyncio.run_coroutine_threadsafe(
            await_awaitable(),
            loop=self._server.loop,
        ).result()

    def await_sync(self, awaitable: Awaitable[T]) -> T:
        self._view_runtime.state = VIEW_RUNTIME_STATE.WAITING_FOR_IOLOOP

        try:
            return self._await_sync(awaitable)

        finally:
            self._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    # html ####################################################################
    def show(
            self,
            html: H = None,
            template: None | str = None,
            template_string: None | str = None,
            title: None | str = None,
            template_context: None | dict = None,
    ) -> None:
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        # templating
        if template or template_string:
            template_context = template_context or {}

            if 'template_context' in template_context:
                template_context = template_context['template_context']

            # string based templates
            if template_string:
                html = self._server.render_string(
                    template_string=template_string,
                    template_context=template_context,
                )

            # file based templates
            else:
                html = self._server.render_template(
                    template_name=cast(str, template),
                    template_context=template_context,
                )

        with self._view_runtime.document.lock:
            html = html or self._view_runtime.document.html

            title, data_type, data = self._view_runtime.document.apply(
                title=title,
                html=html,
            )

            if data_type and data:
                self._view_runtime.send_data(
                    title=title,
                    data=[data_type, data],
                )

    def set_title(self, title: str) -> None:
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        title, _, _ = self._view_runtime.document.apply(title=title)

        with self._view_runtime.document.lock:
            self._view_runtime.send_data(title=title)

    # messaging ###############################################################
    def send_str(
            self,
            string: str,
            broadcast: bool = False,
            filter_connections: Callable[[Connection], bool] = lambda c: True,
            wait: bool = True,
    ) -> None:
        self._assert_not_main_thread()
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        # broadcast
        if broadcast:
            for connection in self.server._websocket_connections.copy():
                if not filter_connections(connection):
                    continue

                connection.send_str(string, wait=wait)

        # view local
        else:
            runtime = self._view_runtime

            with runtime.document.lock:
                for connection in list(runtime.connections.keys()):
                    if not connection.interactive:
                        continue

                    connection.send_str(string, wait=wait)

    # input events ############################################################
    def _await_specific_input_event(
            self,
            *nodes: AbstractNode | list[AbstractNode],
            event_type: str = '',
            html: H = None,
    ) -> InputEvent:
        self._view_runtime.state = VIEW_RUNTIME_STATE.WAITING_FOR_INPUT

        try:
            self._assert_not_main_thread()
            self._assert_view_is_interactive()
            self._assert_view_is_running()

            if len(nodes) == 1 and isinstance(nodes[0], list):
                nodes = tuple(nodes[0])

            nodes = cast('tuple[AbstractNode]', nodes)

            if html is not None:
                self.show(html=html)

            return self._view_runtime.await_input_event(
                nodes=nodes,
                event_type=event_type,
            )

        finally:
            self._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    @overload
    def await_input_event(self, *nodes: AbstractNode, html: H = None) -> InputEvent:  # NOQA: LN001
        ...

    @overload
    def await_input_event(self, __nodes: list[AbstractNode], html: H = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_input_event(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='event',
            html=html,
        )

    @overload
    def await_click(self, *nodes: AbstractNode, html: H = None) -> InputEvent:
        ...

    @overload
    def await_click(self, __nodes: list[AbstractNode], html: H = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_click(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='click',
            html=html,
        )

    @overload
    def await_change(self, *nodes: AbstractNode, html: H = None) -> InputEvent:
        ...

    @overload
    def await_change(self, __nodes: list[AbstractNode], html: H = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_change(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='change',
            html=html,
        )

    @overload
    def await_focus(self, *nodes: AbstractNode, html: H = None) -> InputEvent:
        ...

    @overload
    def await_focus(self, __nodes: list[AbstractNode], html: H = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_focus(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='focus',
            html=html,
        )

    @overload
    def await_blur(self, *nodes: AbstractNode, html: H = None) -> InputEvent:
        ...

    @overload
    def await_blur(self, __nodes: list[AbstractNode], html: H = None) -> InputEvent:  # NOQA: LN001
        ...

    def await_blur(self, *nodes, html=None):
        return self._await_specific_input_event(
            *nodes,
            event_type='blur',
            html=html,
        )

    # view events #############################################################
    def fire_view_event(self, name: str, data: dict | None = None) -> None:
        self.server._view_runtime_controller.fire_view_event(
            name=name,
            data=data,
            view_classes=[self.__class__],
        )

    # runtime #################################################################
    @overload
    def sleep(self, delay: float) -> None:
        ...

    @overload
    def sleep(self, delay: float, result: T) -> T:
        ...

    def sleep(self, delay: float, result: None | T = None) -> None | T:
        self._view_runtime.state = VIEW_RUNTIME_STATE.SLEEPING

        try:
            return self._await_sync(asyncio.sleep(delay, result))

        finally:
            self._view_runtime.state = VIEW_RUNTIME_STATE.RUNNING

    def daemonize(self) -> None:
        self._assert_view_is_interactive()

        self._view_runtime.is_daemon = True

    def ping(self) -> Literal['pong']:
        self._assert_view_is_interactive()
        self._assert_view_is_running()

        return 'pong'

    # hooks ###################################################################
    def handle_request(self, request: Request) -> None | str | AbstractNode | dict:  # NOQA: LN001
        return ''

    def handle_input_event_root(
            self,
            input_event: InputEvent,
    ) -> InputEvent | dict | None:

        return input_event

    def handle_input_event(
            self,
            input_event: InputEvent,
    ) -> InputEvent | dict | None:

        return input_event

    def on_view_event(self, view_event: 'ViewEvent') -> dict | None:
        pass

    def on_stop(self, reason: Exception | None) -> None:
        pass

    def on_cleanup(self) -> None:
        pass



from __future__ import annotations

from typing import Type, cast
import logging
import inspect
import asyncio

from lona.view import LonaView
from lona.routing import Route

logger = logging.getLogger('lona.view_loader')


class ViewLoader:
    # FIXME: type annotations: views can also be callbacks to make
    # aiohttp-wsgi work

    def __init__(self, server):
        self.server = server

        self.setup()

    def _gen_cache_key(
            self,
            view: type[LonaView] | str,
    ) -> type[LonaView] | str | int:

        try:
            hash(view)

            return view

        except TypeError:
            return id(view)

    def _run_checks(self, route: Route, view: type[LonaView]) -> None:
        # check if view is instance of lona.views.LonaView
        if (not route.http_pass_through and
                (not inspect.isclass(view) or not issubclass(view, LonaView))):

            logger.error('%s is no lona view', route.view)

            return

        # check if lona specific hooks are coroutine functions
        hook_names = [
            'handle_request',
            'handle_input_event_root',
            'handle_input_event',
            'on_view_event',
            'on_stop',
            'on_cleanup',
        ]

        for hook_name in hook_names:
            hook = getattr(view, hook_name, None)

            if (not route.http_pass_through and
                    asyncio.iscoroutinefunction(hook)):

                logger.error(
                    '%s.%s is a coroutine function',
                    route.view,
                    hook_name,
                )

    def _generate_acquiring_error_view(
            self,
            exception: Exception,
    ) -> type[LonaView]:

        class AcquiringErrorView(LonaView):
            def handle_request(self, request):
                raise exception

        return AcquiringErrorView

    def _acquire(self, view: type[LonaView] | str) -> type[LonaView]:
        logger.debug('loading %s', view)

        if isinstance(view, str):
            try:
                view = self.server.acquire(view)

            except Exception as exception:
                logger.exception("exception raised while importing '%s'", view)

                view = self._generate_acquiring_error_view(exception)

        return cast('type[LonaView]', view)

    def _cache_view(
            self,
            route: Route | None,
            view: type[LonaView] | str,
    ) -> None:

        view_class = self._acquire(view)

        if route:
            self._run_checks(route, view_class)

        cache_key = self._gen_cache_key(view)
        self._cache[cache_key] = view_class

    def setup(self):
        # views
        logger.debug('loading views from routes')

        self._cache = {}

        for route in self.server._router.routes:
            self._cache_view(
                route=route,
                view=route.view,
            )

            if route.frontend_view:
                self._cache_view(
                    route=None,
                    view=route.frontend_view,
                )

        # special views
        import_strings = [
            # frontend
            self.server.settings.CORE_FRONTEND_VIEW,
            self.server.settings.FRONTEND_VIEW,

            # error 403
            self.server.settings.CORE_ERROR_403_VIEW,
            self.server.settings.ERROR_403_VIEW,

            # error 404
            self.server.settings.CORE_ERROR_404_VIEW,
            self.server.settings.ERROR_404_VIEW,

            # error 500
            self.server.settings.CORE_ERROR_500_VIEW,
            self.server.settings.ERROR_500_VIEW,
        ]

        for import_string in import_strings:
            if not import_string:
                continue

            self._cache_view(
                route=None,
                view=import_string,
            )

    def load(self, view: type[LonaView] | str) -> type[LonaView]:
        cache_key = self._gen_cache_key(view)

        return cast(Type[LonaView], self._cache[cache_key])

    def get_all_views(self) -> list[type[LonaView]]:
        return list(self._cache.values())



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from lona.view import LonaView


@dataclass
class ViewEvent:
    name: str
    data: dict
    view_classes: list[type[LonaView]]



from concurrent.futures import ThreadPoolExecutor


class WorkerPool:
    def __init__(self, settings):
        self.settings = settings

        self._executors = {
            'worker': ThreadPoolExecutor(
                max_workers=self.settings.MAX_WORKER_THREADS,
                thread_name_prefix='LonaWorker',
            ),
            'runtime_worker': ThreadPoolExecutor(
                max_workers=self.settings.MAX_RUNTIME_THREADS,
                thread_name_prefix='LonaRuntimeWorker',
            ),
            'static_worker': None,
        }

        if (self.settings.MAX_STATIC_THREADS and
                self.settings.STATIC_FILES_SERVE):

            self._executors['static_worker'] = ThreadPoolExecutor(
                max_workers=self.settings.MAX_STATIC_THREADS,
                thread_name_prefix='LonaStaticWorker',
            )

    def get_executor(self, name):
        if name not in self._executors:
            raise RuntimeError(f"no executor named '{name}'")

        if not self._executors[name]:
            raise RuntimeError(f"executor '{name}' has no threads setup")

        return self._executors[name]

    def shutdown(self):
        for executor in self._executors.values():
            if not executor:
                continue

            executor.shutdown()



import sys
import os


def terminal_supports_colors():
    return (
        # check if stdout is a tty
        hasattr(sys.stdout, 'isatty') and
        sys.stdout.isatty()
    ) and (
        # Windows checks
        sys.platform != 'win32' or
        'ANSICON' in os.environ or

        # Windows Terminal supports VT codes
        'WT_SESSION' in os.environ or

        # Microsoft Visual Studio Code's built-in terminal supports colors
        os.environ.get('TERM_PROGRAM') == 'vscode'
    )


def colors_are_enabled():
    # https://no-color.org/

    return 'NO_COLOR' not in os.environ



from __future__ import annotations

from typing import overload, Callable, TypeVar, cast, Any
from concurrent.futures import Future
from collections.abc import Awaitable
from functools import reduce
from copy import deepcopy
import contextlib
import operator
import logging
import inspect
import asyncio
import os

from aiohttp.web import (
    WebSocketResponse,
    FileResponse,
    Application,
    HTTPFound,
    Response,
)
from typing_extensions import Literal
from aiohttp import WSMsgType
from jinja2 import Template

from lona.view_runtime_controller import ViewRuntimeController
from lona.middleware_controller import MiddlewareController
from lona.static_file_loader import StaticFileLoader
from lona.response_parser import ResponseParser
from lona.templating import TemplatingEngine
from lona.imports import acquire as _acquire
from lona.view_loader import ViewLoader
from lona.routing import Router, Route
from lona.connection import Connection
from lona.settings import Settings
from lona.request import Request
from lona.protocol import METHOD
from lona.view import LonaView
from lona.state import State

DEFAULT_SETTINGS = os.path.join(
    os.path.dirname(__file__),
    'default_settings.py',
)

server_logger = logging.getLogger('lona.server')
http_logger = logging.getLogger('lona.server.http')
websockets_logger = logging.getLogger('lona.server.websockets')

T = TypeVar('T')


class LonaServer:
    def __init__(self, project_root, settings_paths=None,
                 settings_pre_overrides=None, settings_post_overrides=None,
                 routes=None):

        self._project_root = os.path.abspath(project_root)

        self._websocket_connections = []
        self._loop = None
        self._worker_pool = None
        self._app: Application = None

        # setup settings
        server_logger.debug('setup settings')

        self._settings_paths = [
            DEFAULT_SETTINGS,
        ]

        for path in settings_paths or []:
            self._settings_paths.append(
                os.path.normpath(
                    os.path.join(self.project_root, path),
                ),
            )

        self.settings = Settings()

        if settings_pre_overrides:
            server_logger.debug('applying settings pre overrides')

            self.settings.update(settings_pre_overrides)

        for import_string in self._settings_paths:
            server_logger.debug("loading settings from '%s'", import_string)

            self.settings.add(import_string)

        if settings_post_overrides:
            server_logger.debug('applying settings post overrides')

            self.settings.update(settings_post_overrides)

        # setup aiohttp app
        server_logger.debug("starting server in '%s'", project_root)

        self._app = Application(
            client_max_size=self.settings.AIOHTTP_CLIENT_MAX_SIZE,
        )
        self._app['lona_server'] = self

        self._app.on_startup.append(self._start)
        self._app.on_shutdown.append(self._stop)

        # setup templating
        server_logger.debug('setup templating')

        self._templating_engine = TemplatingEngine(self)

        # setup state
        server_logger.debug('setup state')
        self._state = State(
            initial_data=deepcopy(self.settings.INITIAL_SERVER_STATE),
        )

        # setup routing
        server_logger.debug('setup routing')
        self._router = Router()

        if not routes:
            server_logger.debug("loading routing table from '%s'",
                                self.settings.ROUTING_TABLE)

            routes = self.acquire(self.settings.ROUTING_TABLE)

        if routes:
            self._router.add_routes(*routes)

        else:
            server_logger.warning('routing table is empty')

        # setup middleware controller
        server_logger.debug('setup middleware controller')

        self._middleware_controller = MiddlewareController(self)

        self._app.on_startup.append(
            self._middleware_controller.run_on_startup,
        )

        self._app.on_shutdown.append(
            self._middleware_controller.run_on_shutdown,
        )

        # setup aiohttp routes
        server_logger.debug('setup aiohttp routing')

        static_url = self.settings.STATIC_URL_PREFIX + '{path:.*}'

        server_logger.debug('static url set to %r', static_url)

        self._app.router.add_route(
            '*', static_url, self._handle_static_file_request)

        self._app.router.add_route(
            '*', '/{path_info:.*}', self._handle_http_request)

        # setup view loader
        server_logger.debug('setup view loader')

        self._view_loader = ViewLoader(self)

        # setup response parser
        server_logger.debug('setup response parser')

        self._response_parser = ResponseParser(self)

        # setup views
        server_logger.debug('setup view runtime controller')

        self._view_runtime_controller = ViewRuntimeController(self)

        # setup static files

        # the static file loader has to be started last because it does
        # node class discovery which has to happen after all views are imported
        server_logger.debug('setup static file')

        self._static_file_loader = StaticFileLoader(self)

        # finish
        server_logger.debug('setup finish')

    # properties ##############################################################
    @property
    def project_root(self):
        return self._project_root

    @property
    def loop(self):
        return self._loop

    @property
    def worker_pool(self):
        return self._worker_pool

    @property
    def settings_paths(self):
        return tuple(self._settings_paths)

    @property
    def state(self):
        return self._state

    # template dirs
    @property
    def template_dirs(self):
        return tuple(self._templating_engine.template_dirs)

    @template_dirs.setter
    def template_dirs(self, new_value):
        self._templating_engine.template_dirs = new_value

    # static dirs
    @property
    def static_dirs(self):
        return tuple(self._static_file_loader.static_dirs)

    @static_dirs.setter
    def static_dirs(self, new_value):
        self._static_file_loader.static_dirs = new_value

    # start and stop ##########################################################
    async def _start(self, *args, **kwargs):
        server_logger.debug('start')

        if not self._loop:
            raise RuntimeError('loop is not set')

        if not self._worker_pool:
            raise RuntimeError('worker_pool is not set')

        # run checks
        # tests
        test_names = [
            'TEST_VIEW_START_TIMEOUT',
            'TEST_INPUT_EVENT_TIMEOUT',
        ]

        for test_name in test_names:
            if self.settings.get(test_name, False):
                server_logger.warning('%s is enabled', test_name)

        self._view_runtime_controller.start()

    async def _stop(self, *args, **kwargs):
        server_logger.debug('stop')

        await self.run_function_async(self._view_runtime_controller.stop)

        for connection in self._websocket_connections.copy():
            with contextlib.suppress(Exception):
                await connection.websocket.close()

    # connection management ###################################################
    async def _setup_connection(self, http_request, websocket=None):
        connection = Connection(self, http_request, websocket)

        handled, data, middleware = \
            await self._middleware_controller.handle_connection(
                connection,
            )

        if websocket is not None:
            self._websocket_connections.append(connection)

        return connection, (handled, data, middleware)

    def _remove_connection(self, connection):
        self._view_runtime_controller.remove_connection(connection)

        if connection in self._websocket_connections:
            self._websocket_connections.remove(connection)

    # view helper #############################################################
    def _render_response(self, response_dict):
        if response_dict['file']:
            return FileResponse(response_dict['file'])

        if response_dict['redirect']:
            return HTTPFound(response_dict['redirect'])

        if response_dict['http_redirect']:
            return HTTPFound(response_dict['http_redirect'])

        default_headers = {
            'Cache-Control': 'no-cache, no-store, must-revalidate',
        }

        headers = response_dict['headers'] or default_headers

        response = Response(
            status=response_dict['status'],
            content_type=response_dict['content_type'],
            text=response_dict['text'],
            body=response_dict['body'],
            headers=headers,
        )

        return response

    # handle http requests ####################################################
    async def _handle_static_file_request(self, request):
        def _404():
            return Response(
                status=404,
                text='404: Not found',
            )

        if not self.settings.STATIC_FILES_SERVE:
            server_logger.warning(
                'Reverse proxy seems to be misconfigured: a static file request was received but STATIC_FILES_SERVE is disabled',
            )

            return _404()

        rel_path = request.match_info['path']

        abs_path = await self.run_function_async(
            self._static_file_loader.resolve_path,
            rel_path,
            executor_name='static_worker',
        )

        if not abs_path:
            return _404()

        return FileResponse(abs_path)

    async def _handle_websocket_message(self, connection, message):
        websockets_logger.debug(
            '%s message received %s', connection, message.data)

        handled, data, middleware = \
            await self._middleware_controller.handle_websocket_message(
                connection,
                message.data,
            )

        if not handled:
            websockets_logger.debug(
                '%s message got not handled', connection)

    async def _handle_websocket_request(self, http_request):
        websocket = None
        connection = None

        async def close_websocket():
            await self.run_function_async(
                self._remove_connection,
                connection=connection,
            )

            await websocket.close()

            websockets_logger.debug('%s closed', connection)

        # setup websocket
        websocket = WebSocketResponse()
        await websocket.prepare(http_request)

        # setup connection
        connection, middleware_data = await self._setup_connection(
            http_request,
            websocket,
        )

        handled, data, middleware = middleware_data

        # connection got closed by middleware
        if handled:
            if data:
                if not isinstance(data, str):
                    raise RuntimeError(
                        f'{middleware}.handle_connection returned non string data',
                    )

                await connection.send_str(data, wait=False)

            await close_websocket()

            return websocket

        websockets_logger.debug('%s opened', connection)

        # main loop
        try:
            async for message in websocket:
                if message.type == WSMsgType.TEXT:
                    await self._handle_websocket_message(connection, message)

                elif message.type == WSMsgType.PING:
                    await websocket.pong()

                elif message.type in (WSMsgType.CLOSED, WSMsgType.ERROR):
                    break

        except asyncio.CancelledError:
            websockets_logger.debug('CancelledError')

        finally:
            await close_websocket()

        return websocket

    async def _handle_http_request(self, http_request):
        http_logger.debug('http request incoming')

        # resolve path
        http_logger.debug('resolving path %r', http_request.path)

        match, route, match_info = await self.run_function_async(
            self._router.resolve,
            http_request.path,
        )

        if match:
            http_logger.debug('route %s matched', route)

        else:
            http_logger.debug('no route matched')

        # http pass through ###################################################
        if match and route.http_pass_through:
            http_logger.debug('http_pass_through mode')

            # load view
            view = self._view_loader.load(route.view)

            if inspect.isclass(view):
                view = view(
                    server=self,
                    view_runtime=None,
                    request=http_request,
                ).handle_request

            # run view
            if asyncio.iscoroutinefunction(view):
                response = await view(http_request)

            else:
                response = await self.run_function_async(
                    view,
                    http_request,
                    executor_name='runtime_worker',
                )

            if asyncio.iscoroutine(response):
                response = await response

            # render and return response
            if isinstance(response, dict):
                return await self.run_function_async(
                    self._render_response,
                    response,
                )

            return response

        # websocket requests ##################################################
        if (http_request.method == 'GET' and
                (http_request.headers.get('upgrade', '').lower() ==
                    'websocket')):

            return await self._handle_websocket_request(http_request)

        # non interactive view or frontend ####################################
        # setup connection
        try:
            connection, middleware_data = await self._setup_connection(
                http_request,
            )

        except Exception:
            http_logger.exception(
                'Exception occurred while setting connection up',
            )

            return Response(status=500, body='500: Internal Error')

        handled, data, middleware = middleware_data

        # connection got closed by middleware
        if handled:
            if data is not None:
                if isinstance(data, dict):
                    return self._render_response(data)

                return data

            return Response(status=503, body='503: Service Unavailable')

        post_data = await http_request.post()

        response_dict = await self.run_function_async(
            self._view_runtime_controller.handle_view_message,
            connection=connection,
            window_id=None,
            view_runtime_id=None,
            method=METHOD.VIEW,
            payload=[str(http_request.url), post_data],
            executor_name='runtime_worker',
        )

        return self._render_response(response_dict)

    # public api ##############################################################
    @overload
    def run_coroutine_sync(
            self,
            coroutine: Awaitable[T],
            wait: None | Literal[True] = True,
    ) -> T:
        ...

    @overload
    def run_coroutine_sync(
            self,
            coroutine: Awaitable[T],
            wait: Literal[False],
    ) -> Future[T]:
        ...

    def run_coroutine_sync(
            self,
            coroutine: Awaitable[T],
            wait: None | bool = True,
    ) -> Future[T] | T:

        future = asyncio.run_coroutine_threadsafe(coroutine, loop=self._loop)

        if wait:
            return future.result()

        return future

    def run_function_async(
            self,
            function: Callable,
            *args: Any,
            executor_name: str = 'worker',
            **kwargs: Any,
    ) -> Future:

        """
        Takes a function or callable, runs it asynchronously in an thread
        executor and returns an awaitable future for the result of the given
        function.

        :function:       (callable) callable
        :args:           (list|tuple) args for the callable
        :executor_name:  (str) name of the worker pool (default is 'worker')
        :kwargs:         (dict) keyword argument for the callable
        """

        def _function():
            return function(*args, **kwargs)

        return cast(
            Future,
            self._loop.run_in_executor(
                self._worker_pool.get_executor(executor_name),
                _function,
            ),
        )

    def resolve_path(
            self,
            path: str,
    ) -> str:

        """
        Takes a path as a string and resolves it relatively to
        settings.PROJECT_ROOT.
        If the path starts with '/' the path gets returned unresolved.

        :path: (str) path
        """

        if path.startswith('/'):
            return path

        return os.path.normpath(os.path.join(self.project_root, path))

    def acquire(
            self,
            import_string: str,
            ignore_import_cache: bool = False,
    ) -> Any:

        """
        Takes an import string and returns the imported value.
        if ignore_import_cache is set, the attribute gets imported around
        Pythons import cache.

        Examples:
         - 'foo.bar.baz' (imports 'baz' from 'foo.bar')
         - 'foo/bar.py::baz' (imports 'baz' from 'foo/bar.py')

        :import_string:        (str) import string
        :ignore_import_cache:  (bool) ignore import cache as bool
        """

        if '::' in import_string:
            script, attribute_name = import_string.split('::')
            script = self.resolve_path(script)

            import_string = f'{script}::{attribute_name}'

        return _acquire(import_string, ignore_import_cache=ignore_import_cache)

    def get_running_views_count(
            self,
            user: Any,
    ) -> int:

        """
        Returns the count of running views for the given user as integer.
        """

        return cast(
            int,
            self._view_runtime_controller.get_running_views_count(user),
        )

    def view_is_already_running(
            self,
            request: Request,
    ) -> bool:

        """
        Checks if a view for the given request is already running and returns
        a boolean.

        :request:  (lona.request.Request)
        """

        view_runtime = self._view_runtime_controller.get_running_view_runtime(
            user=request.user,
            route=request.route,
            match_info=request.match_info,
        )

        return bool(view_runtime)

    def get_connection_count(
            self,
            user: Any,
    ) -> int:

        """
        Returns the count of connections for the given user as integer.

        :user: user object
        """

        count = 0

        for connection in self._websocket_connections.copy():
            if connection.user == user:
                count += 1

        return count

    def get_connected_user_count(self) -> int:
        """
        Returns the count of connected users as integer.
        """

        user: list[Any] = []

        def add_user(new_user):
            for _user in user:
                if new_user == _user:
                    return

            user.append(new_user)

        for connection in self._websocket_connections.copy():
            add_user(connection.user)

        return len(user)

    def get_template(
            self,
            template_name: str,
    ) -> Template:

        """
        Returns a Jinja2 template object associated with the given template
        name.
        """

        return cast(
            Template,
            self._templating_engine.get_template(template_name),
        )

    def render_string(
            self,
            template_string: str,
            template_context: dict | None = None,
    ) -> str:

        """
        Takes a Jinja2 template as a string and returns the rendering result
        as string. If no template context is given, an empty one is used.

        All values in settings.TEMPLATE_EXTRA_CONTEXT get added to the
        template context before rendering.

        :template_string:  (str) Jinja2 template as string
        :template_context: (dict|none) template context as dictionary
        """

        return cast(
            str,
            self._templating_engine.render_string(
                template_string=template_string,
                template_context=template_context,
            ),
        )

    def render_template(
            self,
            template_name: str,
            template_context: dict | None = None,
    ) -> str:

        """
        Takes the name of a Jinja2 template as a string, searches for a
        matching template in settings.TEMPLATE_DIRS and returns the rendering
        result as string. If no template context is given, an empty
        one is used.

        All values in settings.TEMPLATE_EXTRA_CONTEXT get added to the
        template context before rendering.

        :template_name:    (str) Jinja2 template as string
        :template_context: (dict|none) template context as dictionary
        """

        return cast(
            str,
            self._templating_engine.render_template(
                template_name=template_name,
                template_context=template_context,
            ),
        )

    def get_view_class(
            self,
            route_name: str | None = None,
            route: Route | None = None,
            import_string: str | None = None,
            url: str | None = None,
    ) -> type[LonaView] | None:

        """
        Returns the lona.view.LonaView subclass associated with the given
        route_name, route, import string or url.

        Only one argument can be set at a time.
        """

        args = [bool(route_name), bool(route), bool(import_string), bool(url)]

        if reduce(operator.xor, args, True):
            raise ValueError('too many or too few arguments given')

        if route_name:
            route = self._router.get_route(name=route_name)

            if not route:
                return None

        if route:
            view = route.view

        if import_string:
            view = import_string

        if url:
            success, route, match_info = self._router.resolve(url)

            if not success:
                return None

            route = cast(Route, route)
            view = route.view

        return cast(
            type[LonaView],
            self._view_loader.load(view),
        )

    def get_views(
            self,
            route_name: str | None = None,
            route: Route | None = None,
            import_string: str | None = None,
            url: str | None = None,
            user: Any = None,
    ) -> list[LonaView]:

        """
        Returns a list of all running Lona views associated with the given
        route_name, route, import string or url.

        Only one argument to find the view can be set at a time. User can
        always be set.

        :route:          (lona.Route|none) Lona route associated with the view
        :import_string:  (str|none) import string of the view
        :url:            (str|none) URL to the view
        :user:           (any) User that runs th view (optional)
        """

        view_class = self.get_view_class(
            route_name=route_name,
            route=route,
            import_string=import_string,
            url=url,
        )

        views = cast(
            list[LonaView],
            self._view_runtime_controller.get_views(view_class=view_class),
        )

        if user:
            for view in views.copy():
                if view.request.user is not user:
                    views.remove(view)

        return views

    def reverse(
            self,
            route_name: str,
            *args: Any,
            **kwargs: dict[str, Any],
    ) -> str:

        """
        Returns a routing reverse match as string.

        :route_name:  (str) route name as string
        :args:        route arguments
        :kwargs:      route keyword argument
        """

        return cast(
            str,
            self._router.reverse(
                route_name=route_name,
                *args,
                **kwargs,
            ),
        )

    def fire_view_event(
            self,
            name: str,
            data: dict | None = None,
            view_classes: type[LonaView] | list[type[LonaView]] | None = None,
    ) -> None:

        """
        Sends a view event to all objects of the given LonaView class.
        If view_classes is not set, the event becomes a broadcast event and
        gets send to all view classes.

        :name: has to be a str, data is optional but has to be a dict if set.
        :view_classes: can be a single view class or a list of view classes.

        Examples:

            # broadcast event
            server.fire_view_event('foo', {'foo': 'bar'})

            # event for all view objects behind the URL '/foo'
            server.fire_view_event(
                'foo',
                {'foo': 'bar'},
                view_classes=server.get_view_class(url='/foo'),
            )
        """

        self._view_runtime_controller.fire_view_event(
            name=name,
            data=data,
            view_classes=view_classes,
        )



from lona.protocol import PATCH_TYPE, OPERATION


class AttributeList:
    PATCH_TYPE: PATCH_TYPE

    def __init__(self, node, *args, **kwargs):
        self._node = node
        self._attributes = set(*args, **kwargs)

    # list helper #############################################################
    def add(self, attribute):
        if not isinstance(attribute, (int, bool, float, str)):
            raise ValueError(f'unsupported type: {type(attribute)}')

        with self._node.lock:
            if attribute in self._attributes:
                return

            self._attributes.add(attribute)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.ADD,
                payload=[
                    attribute,
                ],
            )

    def remove(self, attribute):
        with self._node.lock:
            if attribute not in self._attributes:
                return

            self._attributes.remove(attribute)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.REMOVE,
                payload=[
                    attribute,
                ],
            )

    def clear(self):
        with self._node.lock:
            if not self._attributes:
                return

            self._attributes.clear()

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.CLEAR,
                payload=[],
            )

    def toggle(self, attribute):
        with self._node.lock:
            if attribute in self._attributes:
                self.remove(attribute)

            else:
                self.add(attribute)

    def append(self, attribute):
        return self.add(attribute)

    def extend(self, attributes):
        with self._node.lock:
            for attribute in iter(attributes):
                self.add(attribute)

    def __eq__(self, other):
        with self._node.lock:
            if isinstance(other, self.__class__):
                other = other._attributes

            elif not isinstance(other, (list, set, tuple)):
                return False

            return self._attributes == set(other)

    def __len__(self):
        with self._node.lock:
            return len(self._attributes)

    def __bool__(self):
        with self._node.lock:
            return bool(self._attributes)

    def __iter__(self):
        with self._node.lock:
            return self._attributes.__iter__()

    def __contains__(self, attribute):
        with self._node.lock:
            return attribute in self._attributes

    @property
    def _list(self):
        with self._node.lock:
            return sorted(self._attributes)

    # serialization ###########################################################
    def _reset(self, value):
        if not isinstance(value, list):
            raise ValueError(f'unsupported type: {type(value)}')

        for i in value:
            if not isinstance(i, (int, bool, float, str)):
                raise ValueError(f'unsupported type: {type(i)}')

        with self._node.lock:
            self._attributes = set(value)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=self.PATCH_TYPE,
                operation=OPERATION.RESET,
                payload=[
                    list(value),
                ],
            )

    def _serialize(self):
        return self._list

    # string representation ###################################################
    def __str__(self):
        return ' '.join(self._list)

    def __repr__(self):
        return f'<AttributeList({self._list!r})>'


class IDList(AttributeList):
    PATCH_TYPE = PATCH_TYPE.ID_LIST

    def __repr__(self):
        return f'<IDList({self._list!r})>'


class ClassList(AttributeList):
    PATCH_TYPE = PATCH_TYPE.CLASS_LIST

    def __repr__(self):
        return f'<ClassList({self._list!r})>'



try:
    from .exceptions import *  # NOQA: F403
    from .routing import MATCH_ALL, Route
    from .errors import *  # NOQA: F403
    from .view import LonaView
    from .app import LonaApp

except ImportError as e:  # pragma: no cover
    # this can happen while installing the package and can be ignored
    if e.name != 'typing_extensions':
        raise

VERSION = (1, 10, 5)
VERSION_STRING = '.'.join(str(i) for i in VERSION)



from copy import deepcopy

from lona.protocol import PATCH_TYPE, OPERATION


def check_value(value):
    if value is None:
        return

    if not isinstance(value, (bool, int, float, str, list, dict)):
        raise ValueError(f'unsupported type: {type(value)}')

    if isinstance(value, list):
        for i in value:
            check_value(i)

    if isinstance(value, dict):
        for k, v in value.items():
            check_value(k)
            check_value(v)


class ListOverlay:
    def __init__(self, widget_data, key_path, original_data):
        self._widget_data = widget_data
        self._widget = widget_data._widget
        self._key_path = key_path
        self._original_data = original_data

    def append(self, item):
        check_value(item)

        with self._widget.lock:
            self._original_data.append(item)

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.INSERT,
                payload=[
                    self._key_path,
                    len(self._original_data) - 1,
                    deepcopy(item),
                ],
            )

    def clear(self):
        with self._widget.lock:
            self._original_data.clear()

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.CLEAR,
                payload=[
                    self._key_path,
                ],
            )

    def copy(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.copy(*args, **kwargs)

    def count(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.count(*args, **kwargs)

    def extend(self, items):
        check_value(items)

        with self._widget.lock:
            for item in items:
                self.append(item)

    def index(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.count(*args, **kwargs)

    def insert(self, index, item):
        check_value(item)

        with self._widget.lock:
            self._original_data.insert(index, item)

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.INSERT,
                payload=[
                    self._key_path,
                    self._original_data.index(item),
                    deepcopy(item),
                ],
            )

    def pop(self, index):
        with self._widget.lock:
            item = self._original_data.pop(index)

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.REMOVE,
                payload=[
                    self._key_path,
                    index,
                ],
            )

            return item

    def remove(self, item):
        with self._widget.lock:
            index = 0

            for i in self._original_data:
                if i == item:
                    self._widget.document.add_patch(
                        node_id=self._widget.id,
                        patch_type=PATCH_TYPE.WIDGET_DATA,
                        operation=OPERATION.REMOVE,
                        payload=[
                            self._key_path,
                            index,
                        ],
                    )

                index += 1

            self._original_data.remove(item)

    def reverse(self, *args, **kwargs):
        raise NotImplementedError

    def sort(self, *args, **kwargs):
        raise NotImplementedError

    def __setitem__(self, name, item):
        with self._widget.lock:
            check_value(name)
            check_value(item)

            self._original_data[name] = item

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.SET,
                payload=[
                    self._key_path,
                    name,
                    deepcopy(item),
                ],
            )

    def __getitem__(self, name):
        with self._widget.lock:
            item = self._original_data[name]

            if isinstance(item, list):
                return ListOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            if isinstance(item, dict):
                return DictOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            return item

    def __delitem__(self, name):
        with self._widget.lock:
            del self._original_data[name]

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.REMOVE,
                payload=[
                    self._key_path,
                    name,
                ],
            )

    def __str__(self, *args, **kwargs):
        with self._widget.lock:
            return str(self._original_data)

    def __repr__(self, *args, **kwargs):
        with self._widget.lock:
            return repr(self._original_data)

    def __len__(self, *args, **kwargs):
        with self._widget.lock:
            return len(self._original_data)

    def __bool__(self, *args, **kwargs):
        with self._widget.lock:
            return bool(self._original_data)


class DictOverlay:
    def __init__(self, widget_data, key_path, original_data):
        self._widget_data = widget_data
        self._widget = widget_data._widget
        self._key_path = key_path
        self._original_data = original_data

    def clear(self):
        with self._widget.lock:
            self._original_data.clear()

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.CLEAR,
                payload=[
                    self._key_path,
                ],
            )

    def copy(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.copy(*args, **kwargs)

    def fromkeys(self, *args, **kwargs):
        raise NotImplementedError

    def get(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.get(*args, **kwargs)

    def items(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.items()

    def keys(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.keys()

    def pop(self, key):
        with self._widget.lock:
            item = self._original_data.pop(key)

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.REMOVE,
                payload=[
                    self._key_path,
                    key,
                ],
            )

            return item

    def popitem(self):
        with self._widget.lock:
            key, value = self._original_data.popitem()

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.REMOVE,
                payload=[
                    self._key_path,
                    key,
                ],
            )

            return key, value

    def setdefault(self, *args, **kwargs):
        raise NotImplementedError

    def update(self, update_dict):
        with self._widget.lock:
            for key, value in update_dict.items():
                self._original_data[key] = value

                self._widget.document.add_patch(
                    node_id=self._widget.id,
                    patch_type=PATCH_TYPE.WIDGET_DATA,
                    operation=OPERATION.SET,
                    payload=[
                        self._key_path,
                        key,
                        deepcopy(value),
                    ],
                )

    def values(self, *args, **kwargs):
        with self._widget.lock:
            return self._original_data.values()

    def __setitem__(self, name, item):
        with self._widget.lock:
            check_value(name)
            check_value(item)

            self._original_data[name] = item

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.SET,
                payload=[
                    self._key_path,
                    name,
                    deepcopy(item),
                ],
            )

    def __getitem__(self, name):
        with self._widget.lock:
            item = self._original_data[name]

            if isinstance(item, list):
                return ListOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            if isinstance(item, dict):
                return DictOverlay(
                    widget_data=self._widget_data,
                    key_path=self._key_path + [name],
                    original_data=item,
                )

            return item

    def __delitem__(self, name):
        with self._widget.lock:
            del self._original_data[name]

            self._widget.document.add_patch(
                node_id=self._widget.id,
                patch_type=PATCH_TYPE.WIDGET_DATA,
                operation=OPERATION.REMOVE,
                payload=[
                    self._key_path,
                    name,
                ],
            )

    def __str__(self, *args, **kwargs):
        with self._widget.lock:
            return str(self._original_data)

    def __repr__(self, *args, **kwargs):
        with self._widget.lock:
            return repr(self._original_data)

    def __len__(self, *args, **kwargs):
        with self._widget.lock:
            return len(self._original_data)

    def __bool__(self, *args, **kwargs):
        with self._widget.lock:
            return bool(self._original_data)


class WidgetData:
    _LONA_CLASS_NAME = 'WidgetData'

    def __init__(self, widget):
        self._widget = widget

        self._reset({}, initial=True)

    def __getitem__(self, *args, **kwargs):
        return self._overlay.__getitem__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        return self._overlay.__setitem__(*args, **kwargs)

    def __delitem__(self, *args, **kwargs):
        return self._overlay.__delitem__(*args, **kwargs)

    def __len__(self, *args, **kwargs):
        with self._widget.lock:
            return len(self._original_data)

    def __bool__(self, *args, **kwargs):
        with self._widget.lock:
            return bool(self._original_data)

    def __getattribute__(self, name):
        if name.startswith('_'):
            return super().__getattribute__(name)

        if hasattr(self._overlay, name):
            return getattr(self._overlay, name)

        return super().__getattribute__(name)

    def __dir__(self):
        return [
            *super().__dir__(),
            *dir(self._overlay),
        ]

    # serialization ###########################################################
    def _reset(self, value, initial=False):
        if not isinstance(value, (dict, list)):
            raise ValueError('widget state has to be dict or list')

        check_value(value)

        with self._widget.lock:
            self._data = value

            if not initial:
                self._widget.document.add_patch(
                    node_id=self._widget.id,
                    patch_type=PATCH_TYPE.WIDGET_DATA,
                    operation=OPERATION.RESET,
                    payload=[
                        [],
                        deepcopy(value),
                    ],
                )

            if isinstance(value, list):
                self._overlay = ListOverlay(
                    widget_data=self,
                    key_path=[],
                    original_data=self._data,
                )

            elif isinstance(value, dict):
                self._overlay = DictOverlay(
                    widget_data=self,
                    key_path=[],
                    original_data=self._data,
                )

    def _serialize(self):
        return deepcopy(self._data)

    # string representation ###################################################
    def __repr__(self):
        return f'<WidgetData({self._data!r}))>'



from lona.html.data_binding.select import *  # NOQA: F403
from lona.html.data_binding.inputs import *  # NOQA: F403
from lona.html.parsing import _setup_node_classes_cache
from lona.events.event_types import *  # NOQA: F403
from lona.html.nodes import *  # NOQA: F403
from lona.html.widget import Widget
from lona.html.widgets import HTML

_setup_node_classes_cache()



import __main__  # isort: skip

from pprint import pformat
import traceback
import textwrap
import inspect
import sys

from rlpython.utils.attribute_table import write_attribute_table
from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table

from lona.view_runtime import VIEW_RUNTIME_STATE


class LonaViewsCommand:
    """
    List currently running Lona views
    """

    NAME = 'lona_views'

    def __init__(self, repl):
        self.repl = repl

    def complete(self, text, state, line_buffer):
        server = self.repl.globals['server']
        controller = server._view_runtime_controller

        view_runtime_ids = []

        for view_runtime in controller.iter_view_runtimes():
            view_runtime_ids.append(view_runtime.view_runtime_id)

        view_runtime_ids = sorted(view_runtime_ids)
        candidates = []

        for view_runtime_id in view_runtime_ids:
            if view_runtime_id.startswith(text):
                candidates.append(view_runtime_id)

        candidates.append(None)

        return candidates[state]

    # helper ##################################################################
    def _get_view_path(self, view_runtime):
        if view_runtime.view.__module__ == '__main__':
            return __main__.__file__

        if view_runtime.view.__module__.endswith('.py'):
            return view_runtime.view.__module__

        return inspect.getabsfile(view_runtime.view.__class__)

    def _get_view_name(self, view_runtime):
        return view_runtime.view.__class__.__name__

    def _print_exception(self, exception):
        lines = traceback.format_exception(
            etype=type(exception),
            value=exception,
            tb=exception.__traceback__,
        )

        for line in lines:
            self.repl.write(line)

    def _get_shortened_stack(self, thread_ident, path):
        frame = sys._current_frames()[thread_ident]

        # frame summary list
        frame_summary_list = traceback.extract_stack(frame)

        while len(frame_summary_list) > 0:
            if frame_summary_list[0].filename == path:
                break

            frame_summary_list.pop(0)

        while len(frame_summary_list) > 0:
            if frame_summary_list[-1].filename == path:
                break

            frame_summary_list.pop(-1)

        # frame list
        frame_list = []

        while frame.f_back is not None:
            frame_list.insert(0, frame)
            frame = frame.f_back

        while len(frame_list) > 0:
            if frame_list[0].f_code.co_filename == path:
                break

            frame_list.pop(0)

        while len(frame_list) > 0:
            if frame_list[-1].f_code.co_filename == path:
                break

            frame_list.pop(-1)

        return frame_summary_list, frame_list

    # command #################################################################
    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='lona_views')

        argument_parser.add_argument('runtime-id', nargs='?')
        argument_parser.add_argument('-m', '--memory', action='store_true')

        arguments = vars(argument_parser.parse_args(argv[1:]))

        if arguments['memory']:
            return self.show_view_memory(arguments)

        if arguments['runtime-id']:
            return self.show_view_info(arguments)

        return self.list_views(arguments)

    def show_view_memory(self, arguments):
        server = self.repl.globals['server']
        controller = server._view_runtime_controller

        if not arguments['runtime-id']:
            self.repl.write_error('no runtime id given')

            return 1

        try:
            view_runtime = controller.get_view_runtime(
                view_runtime_id=arguments['runtime-id'],
            )

        except Exception:
            view_runtime = None

        if not view_runtime:
            self.repl.write_error('invalid runtime id\n')

            return 1

        path = self._get_view_path(view_runtime)

        frame_summary_list, frame_list = \
            self._get_shortened_stack(view_runtime.thread_ident, path)

        self.repl.write('View Memory\n')
        self.repl.write('===========\n\n')

        for index, frame_summary in enumerate(frame_summary_list):
            frame = frame_list[index]

            for line in traceback.format_list([frame_summary]):
                self.repl.write(textwrap.dedent(line))

            rows = [['Key', 'Value']]

            for key, value in frame.f_locals.items():
                rows.append([key, pformat(value)])

            write_table(rows, self.repl.write)
            self.repl.write('\n\n')

    def show_view_info(self, arguments):
        server = self.repl.globals['server']
        controller = server._view_runtime_controller

        try:
            view_runtime = controller.get_view_runtime(
                view_runtime_id=arguments['runtime-id'],
            )

        except Exception:
            view_runtime = None

        if not view_runtime:
            self.repl.write_error('invalid runtime id\n')

            return 1

        # runtime info
        self.repl.write('Runtime Info\n')
        self.repl.write('============\n')

        path = self._get_view_path(view_runtime)
        name = self._get_view_name(view_runtime)

        rows = [
            ['Thread ID', view_runtime.thread_ident],
            ['Thread Name', view_runtime.thread_name],
            ['View ID', view_runtime.view_runtime_id],
            ['Daemon', repr(view_runtime.is_daemon)],
            ['Started at', str(view_runtime.started_at)],
            ['Stopped at', str(view_runtime.stopped_at or '-')],
            ['State', view_runtime.state.name],
        ]

        write_attribute_table(rows, self.repl.write)
        self.repl.write('\n')

        # view info
        self.repl.write('View Info\n')
        self.repl.write('=========\n')

        rows = [
            ['Name', name],
            ['Path', path],
            ['User', repr(view_runtime.request.user)],
            ['Route', repr(view_runtime.route)],
            ['Match info', repr(view_runtime.match_info)],
            ['Method', repr(view_runtime.request.method)],
            ['GET Data', pformat(view_runtime.request.GET)],
            ['POST Data', pformat(view_runtime.request.POST)],
        ]

        write_attribute_table(rows, self.repl.write)
        self.repl.write('\n')

        # connections
        self.repl.write('Connections\n')
        self.repl.write('===========\n')

        rows = [['User', 'Connection', 'Window ID', 'URL']]

        for connection, (window_id, url) in view_runtime.connections.items():
            rows.append([
                repr(connection.user),
                repr(connection),
                window_id,
                repr(url),
            ])

        write_attribute_table(rows, self.repl.write)
        self.repl.write('\n')

        # crash report
        if view_runtime.state == VIEW_RUNTIME_STATE.CRASHED:
            self.repl.write('Crash Report\n')
            self.repl.write('============\n')

            self._print_exception(view_runtime.stop_reason)

        # shortened stack
        elif (view_runtime.state.value > 20 and
                view_runtime.state.value < 30):

            self.repl.write('Shortened Stack\n')
            self.repl.write('===============\n')

            frame_summary_list, frame_list = \
                self._get_shortened_stack(view_runtime.thread_ident, path)

            for line in traceback.format_list(frame_summary_list):
                self.repl.write(line)

        self.repl.write('\n')

    def list_views(self, arguments):
        server = self.repl.globals['server']
        controller = server._view_runtime_controller

        rows = [
            ['Runtime ID', 'Thread ID', 'Flags', 'User', 'Route ID',
             'URL', 'State'],
        ]

        for view_runtime in controller.iter_view_runtimes():
            # find route id
            route = view_runtime.route

            if route:
                route_id = server._router.routes.index(view_runtime.route)

                url = route.format_string.format(
                    **view_runtime.request.match_info)

            else:
                # in case of 404 or 500 responses the runtime
                # may have no route set

                route_id = ''
                url = ''

            # flags
            flags = ''

            if isinstance(view_runtime.request.user, list):
                flags += 'M'

            if view_runtime.is_daemon:
                flags += 'D'

            # add table row
            rows.append([
                str(view_runtime.view_runtime_id),
                view_runtime.thread_ident,
                flags,
                repr(view_runtime.request.user),
                route_id,
                url,
                view_runtime.state.name,
            ])

        write_table(rows, self.repl.write)



class ForbiddenError(Exception):
    pass


class NotFoundError(Exception):
    pass


class ClientError(Exception):
    def __init__(self, message, *args, **kwargs):
        self.message = message

        super().__init__(*args, **kwargs)



import re

SELECTOR_RE = re.compile(r'([#.])?(([^#.\[\]]+)|(\[([^=\[\]]+)=([^=\[\]]+)\]))')
UNSUPPORTED_CHARACTERS = re.compile(r'([ $|!~>+])')


class Selector:
    def __init__(self, raw_selector_string):
        self.raw_selector_string = raw_selector_string

        self.check_raw_selector_string()
        self.parse_selector()

    def __repr__(self):
        return f'<{self.__class__.__name__}({self.selectors!r})>'  # pragma: no cover

    def check_raw_selector_string(self):
        match = UNSUPPORTED_CHARACTERS.search(self.raw_selector_string)

        if match:
            raise ValueError(
                f'unsupported selector feature: {match.group()!r}',
            )

    def parse_selector(self):
        self.selectors = []

        selector_strings = self.raw_selector_string.split(',')

        for selector_string in selector_strings:
            selector = {
                'tag_name': '',
                'ids': [],
                'classes': [],
                'attributes': {},
            }

            for match in SELECTOR_RE.findall(selector_string):
                prefix, name, _, _, attribute_name, attribute_value = match

                if prefix == '#':
                    selector['ids'].append(name)

                elif prefix == '.':
                    selector['classes'].append(name)

                elif not prefix and not attribute_name:
                    selector['tag_name'] = name

                elif attribute_name and attribute_value:
                    attribute_value = attribute_value.replace('"', '')
                    attribute_value = attribute_value.replace("'", '')

                    selector['attributes'][attribute_name] = attribute_value

            self.selectors.append(selector)

    def _match_selector(self, node, selector):
        # tag name
        if selector['tag_name']:
            if not hasattr(node, 'tag_name'):
                return False

            if node.tag_name != selector['tag_name']:
                return False

        # ids
        if selector['ids']:
            if not hasattr(node, 'id_list'):
                return False

            for id_name in selector['ids']:
                if id_name not in node.id_list:
                    return False

        # classes
        if selector['classes']:
            if not hasattr(node, 'class_list'):
                return False

            for class_name in selector['classes']:
                if class_name not in node.class_list:
                    return False

        # attributes
        if selector['attributes']:
            if not hasattr(node, 'attributes'):
                return False

            for name, value in selector['attributes'].items():
                if name not in node.attributes:
                    return False

                if node.attributes[name] != value:
                    return False

        return True

    def match(self, node):
        for selector in self.selectors:
            if self._match_selector(node, selector):
                return True

        return False



import logging

from lona.view import LonaView

logger = logging.getLogger('lona.views')


class FallbackView(LonaView):
    VIEW_SETTING = ''
    TEMPLATE_SETTING = ''

    def render_default_template(self, request, **extra_context):
        template_setting = request.server.settings.get(self.TEMPLATE_SETTING)

        return {
            'template': template_setting,
            'request': request,
            **extra_context,
        }

    def handle_request(self, request, **extra_context):
        view_setting = request.server.settings.get(self.VIEW_SETTING, '')

        if view_setting:
            try:
                view_class = request.server._view_loader.load(view_setting)

                view = view_class(
                    server=self.server,
                    view_runtime=self._view_runtime,
                    request=request,
                )

                return view.handle_request(request, **extra_context)

            except Exception:
                logger.exception(
                    "exception raised while running '%s'. running fallback view",
                    view_setting,
                )

                return self.render_default_template(request, **extra_context)

        return self.render_default_template(request, **extra_context)


class FrontendView(FallbackView):
    VIEW_SETTING = 'FRONTEND_VIEW'
    TEMPLATE_SETTING = 'FRONTEND_TEMPLATE'


class Error403View(FallbackView):
    VIEW_SETTING = 'ERROR_403_VIEW'
    TEMPLATE_SETTING = 'ERROR_403_TEMPLATE'


class Error404View(FallbackView):
    VIEW_SETTING = 'ERROR_404_VIEW'
    TEMPLATE_SETTING = 'ERROR_404_TEMPLATE'


class Error500View(FallbackView):
    VIEW_SETTING = 'ERROR_500_VIEW'
    TEMPLATE_SETTING = 'ERROR_500_TEMPLATE'



from lona.html.abstract_node import AbstractNode
from lona.protocol import PATCH_TYPE, OPERATION
from lona.html.text_node import TextNode


class NodeList:
    def __init__(self, node):
        self._node = node
        self._nodes = []

    # list helper #############################################################
    def _check_node(self, node):
        if isinstance(node, (str, int, float, bool)):
            node = TextNode(node)

        if not isinstance(node, AbstractNode):
            raise ValueError(f'unsupported type: {type(node)}')

        return node

    def _prepare_node(self, node):
        if node.parent and node.parent is not self._node:
            node.parent.remove(node)

        elif node is self._node.root:
            raise RuntimeError('loop detected')

        node._set_parent(self._node)

    def insert(self, index, node):
        node = self._check_node(node)

        with self._node.lock:
            self._prepare_node(node)
            self._nodes.insert(index, node)

            index = self._nodes.index(node)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=PATCH_TYPE.NODES,
                operation=OPERATION.INSERT,
                payload=[
                    index,
                    node._serialize(),
                ],
            )

    def append(self, node):
        node = self._check_node(node)

        with self._node.lock:
            self._prepare_node(node)
            self._nodes.append(node)

            index = self._nodes.index(node)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=PATCH_TYPE.NODES,
                operation=OPERATION.INSERT,
                payload=[
                    index,
                    node._serialize(),
                ],
            )

    def extend(self, nodes):
        with self._node.lock:
            for node in nodes:
                self.append(node)

    def remove(self, node):
        with self._node.lock:
            self._nodes.remove(node)

            node._set_parent(None)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=PATCH_TYPE.NODES,
                operation=OPERATION.REMOVE,
                payload=[
                    node.id,
                ],
            )

    def pop(self, index):
        with self._node.lock:
            node = self._nodes.pop(index)

            node._set_parent(None)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=PATCH_TYPE.NODES,
                operation=OPERATION.REMOVE,
                payload=[
                    node.id,
                ],
            )

            return node

    def clear(self):
        with self._node.lock:
            if not self._nodes:
                return

            for node in list(self._nodes):
                node._set_parent(None)
                self._nodes.remove(node)

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=PATCH_TYPE.NODES,
                operation=OPERATION.CLEAR,
                payload=[],
            )

    def index(self, node):
        with self._node.lock:
            return self._nodes.index(node)

    def __getitem__(self, index):
        with self._node.lock:
            return self._nodes[index]

    def __setitem__(self, index, node):
        node = self._check_node(node)

        with self._node.lock:
            self._prepare_node(node)
            self._nodes[index] = node

            self._node.document.add_patch(
                node_id=self._node.id,
                patch_type=PATCH_TYPE.NODES,
                operation=OPERATION.SET,
                payload=[
                    index,
                    node._serialize(),
                ],
            )

    def __eq__(self, other):
        with self._node.lock:
            if isinstance(other, self.__class__):
                other = other._nodes

            elif not isinstance(other, (list, tuple, set)):
                return False

            if not len(self._nodes) == len(other):
                return False

            for index, node in enumerate(self._nodes):
                if node != other[index]:
                    return False

            return True

    def __in__(self, other):
        with self._node.lock:
            return other in self._nodes

    def __bool__(self):
        with self._node.lock:
            return bool(self._nodes)

    def __len__(self):
        with self._node.lock:
            return self._nodes.__len__()

    def __iter__(self):
        with self._node.lock:
            return self._nodes.__iter__()

    def __contains__(self, node):
        with self._node.lock:
            return node in self._nodes

    # serialization ###########################################################
    def _reset(self, value):
        if not isinstance(value, list):
            value = [value]

        with self._node.lock:
            self._nodes.clear()

            for node in value:
                node = self._check_node(node)
                self._prepare_node(node)
                self._nodes.append(node)

                self._node.document.add_patch(
                    node_id=self._node.id,
                    patch_type=PATCH_TYPE.NODES,
                    operation=OPERATION.RESET,
                    payload=[
                        [i._serialize() for i in self._nodes],
                    ],
                )

    def _serialize(self):
        return [i._serialize() for i in self._nodes]

    # string representation ###################################################
    def __str__(self):
        with self._node.lock:
            return '\n'.join([str(i) for i in self._nodes])

    def __repr__(self):
        return f'<NodeList({self._nodes!r}))>'



from pprint import pformat

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table


class LonaServerStateCommand:
    """
    Show current server.state
    """

    NAME = 'lona_server_state'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl, prog='lona_server_state')

        argument_parser.parse_args(argv[1:])

        server = self.repl.globals['server']

        # write table
        rows = [['Key', 'Value']]

        with server.state.lock:
            for key, value in server.state.items():
                rows.append(
                    [pformat(key), pformat(value)],
                )

        if len(rows) == 1:
            return

        write_table(rows, self.repl.write)



from copy import deepcopy
import logging
import types
import runpy

logger = logging.getLogger('lona.settings')


class Settings:
    def __init__(self):
        self._paths = []
        self._values = {}

    def add(self, path):
        if not path.endswith('.py'):
            logger.error(
                "'%s' does not look like python script (modules are not supported)",
                path,
            )

            return

        self._paths.append(path)

        try:
            values = runpy.run_path(
                path,
                init_globals=self._values,
                run_name=path,
            )

        except Exception:
            logger.exception(
                "exception raised while loading settings '%s'",
                path,
            )

            return

        self._values = {}

        for key, value in values.items():
            if key.startswith('_'):
                continue

            if isinstance(value, types.ModuleType):
                continue

            self._values[key] = deepcopy(value)

    def get(self, *args):
        return self._values.get(*args)

    def update(self, settings):
        self._values.update(settings)

    def __iter__(self):
        return self._values.__iter__()

    def __getattribute__(self, name):
        if name in ('get', 'add', 'update') or name.startswith('_'):
            return super().__getattribute__(name)

        return self._values[name]

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return super().__setattr__(name, value)

        self._values[name] = value

    def __dir__(self):
        return [
            *super().__dir__(),
            *self._values.keys(),
        ]



import builtins
import logging
import json
import os

from jinja2 import FileSystemLoader, Environment

from lona.protocol import get_enum_values

logger = logging.getLogger('lona.templating')

BUILTINS = {}

for name in dir(builtins):
    if name.startswith('_'):
        continue

    BUILTINS[name] = getattr(builtins, name)


class Namespace:
    def __init__(self, server, templating_engine):
        self.server = server
        self.templating_engine = templating_engine

    @property
    def reverse(self):
        return self.server.reverse

    @property
    def settings(self):
        return self.server.settings

    @property
    def state(self):
        return self.server.state

    def load_stylesheets(self):
        return self.server._static_file_loader.style_tags_html

    def load_scripts(self):
        return self.server._static_file_loader.script_tags_html

    def _import(self, *args, **kwargs):
        return self.server.acquire(*args, **kwargs)

    def load_static_file(self, path):
        logger.debug('resolving static file path %s', path)

        if path.startswith('/'):
            path = path[1:]

        if path in self.templating_engine.static_path_cache:
            logger.debug('%s is cached', path)

        else:
            resolved_path = self.server._static_file_loader.resolve_path(path)

            if not resolved_path:
                logger.error("static file '%s' was not found", path)

                return ''

            self.templating_engine.static_path_cache.append(path)

        return os.path.join(
            self.server.settings.STATIC_URL_PREFIX,
            path,
        )

    def get_protocol_as_json(self):
        return json.dumps(get_enum_values())

    def get_settings_as_json(self):
        settings = {
            'DEBUG': self.settings.CLIENT_DEBUG,
            'VIEW_START_TIMEOUT': self.settings.CLIENT_VIEW_START_TIMEOUT,
            'INPUT_EVENT_TIMEOUT': self.settings.CLIENT_INPUT_EVENT_TIMEOUT,
            'PING_INTERVAL': self.settings.CLIENT_PING_INTERVAL,
        }

        return json.dumps(settings)

    def __getattribute__(self, name):
        # this is necessary because in python its illegal to name a
        # function 'import'

        if name == 'import':
            return super().__getattribute__('_import')

        return super().__getattribute__(name)


class TemplatingEngine:
    def __init__(self, server):
        self.server = server

        self.template_dirs = (self.server.settings.TEMPLATE_DIRS +
                              self.server.settings.CORE_TEMPLATE_DIRS)

        self.static_path_cache = []

        # resolving potential relative paths
        for index, template_dir in enumerate(self.template_dirs):
            self.template_dirs[index] = self.server.resolve_path(template_dir)

        logger.debug('loading template_dirs %s',
                     repr(self.template_dirs)[1:-1])

        self.jinja2_env = Environment(
            loader=FileSystemLoader(
                self.template_dirs,
                followlinks=True,
            ),
        )

        # load extra filters
        logger.debug(
            'loading extra filters from settings.TEMPLATE_EXTRA_FILTERS',
        )

        for name, func in self.server.settings.TEMPLATE_EXTRA_FILTERS.items():
            self.jinja2_env.filters[name] = func

    # public api ##############################################################
    def get_template(self, template_name):
        logger.debug("searching for '%s'", template_name)

        template = self.jinja2_env.get_template(template_name)

        logger.debug("'%s' is '%s' ", template_name, template.filename)

        return template

    def generate_template_context(self, overrides):
        context = {
            'Lona': Namespace(
                server=self.server,
                templating_engine=self,
            ),
            **BUILTINS,
            **self.server.settings.TEMPLATE_EXTRA_CONTEXT,
            **overrides,
        }

        context['_variables'] = context

        return context

    def render_string(self, template_string, template_context=None):
        template = self.jinja2_env.from_string(template_string)

        template_context = self.generate_template_context(
            overrides=template_context or {})

        return template.render(template_context)

    def render_template(self, template_name, template_context=None):
        template = self.get_template(template_name)

        template_context = self.generate_template_context(
            overrides=template_context or {})

        return template.render(template_context)



from copy import copy

ATTRIBUTE_NAME = 'data-lona-events'


class NodeEventList:
    def __init__(self, node, event_types):
        self._node = node
        self._event_types = copy(event_types)

        self._apply()

    def __repr__(self):
        return f'<NodeEventList({self._event_types})>'

    @property
    def lock(self):
        return self._node.lock

    def _get_index(self, event_type):
        for index, _event_type in enumerate(self._event_types):
            if _event_type == event_type:
                return index

        return -1

    def _apply(self):
        with self.lock:
            attribute_string = []

            for event_type in self._event_types:
                attribute_string.append(event_type.serialize())

            attribute_string = ';'.join(attribute_string)

            if attribute_string:
                self._node.attributes[ATTRIBUTE_NAME] = attribute_string

            elif ATTRIBUTE_NAME in self._node.attributes:
                del self._node.attributes[ATTRIBUTE_NAME]

    def _reset(self, value):
        with self.lock:
            self._event_types = value
            self._apply()

    def add(self, event_type):
        with self.lock:
            index = self._get_index(event_type)

            if index > -1:
                self._event_types.pop(index)

            self._event_types.append(event_type)
            self._apply()

    def remove(self, event_type):
        with self.lock:
            index = self._get_index(event_type)

            if index == -1:
                return

            self._event_types.remove(event_type)
            self._apply()

    def clear(self):
        with self.lock:
            if not self._event_types:
                return

            self._event_types.clear()
            self._apply()

    def append(self, *args, **kwargs):
        return self.add(*args, **kwargs)

    def extend(self, event_types):
        with self.lock:
            for event_type in event_types:
                if event_type in self._event_types:
                    continue

                self._event_types.append(event_type)

            self._apply()

    def toggle(self, event_type):
        with self.lock:
            if event_type in self._event_types:
                self._event_types.remove(event_type)

            else:
                self._event_types.append(event_type)

            self._apply()



from traceback import format_exception
from textwrap import indent
import threading
import datetime
import logging
import socket
import os

from lona.command_line.terminal import (
    terminal_supports_colors,
    colors_are_enabled,
)

try:
    # syslog is only on unix based systems available
    import syslog

    SYSLOG_IS_AVAILABLE = True

except ImportError:
    SYSLOG_IS_AVAILABLE = False


def journald_is_running():
    return 'JOURNAL_STREAM' in os.environ and 'TERM' not in os.environ


def get_syslog_priority(levelno):
    if levelno <= logging.DEBUG:
        return syslog.LOG_DEBUG

    elif levelno <= logging.INFO:
        return syslog.LOG_INFO

    elif levelno <= logging.WARNING:
        return syslog.LOG_WARNING

    elif levelno <= logging.ERROR:
        return syslog.LOG_ERR

    elif levelno <= logging.CRITICAL:
        return syslog.LOG_CRIT

    return syslog.LOG_ALERT


class LogFilter(logging.Filter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.excluded = []
        self.included = []

    def clear(self):
        self.excluded.clear()
        self.included.clear()

    def include(self, logger_name):
        self.included.append(logger_name)

    def exclude(self, logger_name):
        self.excluded.append(logger_name)

    def filter(self, record):
        # filter exceptions that lona.command_line.run_server.run_server
        # handles it self
        if record.exc_info:

            # OSErrors
            if (isinstance(record.exc_info[1], OSError) and
                    record.exc_info[1].errno in (13, 98)):

                return False

            # socket.gaierror
            if (isinstance(record.exc_info[1], socket.gaierror) and
                    record.exc_info[1].errno in (-2, )):

                return False

        if record.name == 'lona':
            # The lona root logger is used by the command line tools
            # to report errors, for example when startup is not possible due
            # an invalid host or port.
            # These errors can't be ignored.

            return True

        if record.name in self.excluded:
            return False

        if self.included and record.name not in self.included:
            return False

        return True


class LogFormatter(logging.Formatter):
    def __init__(self, *args, syslog_priorities=False, **kwargs):
        super().__init__(*args, **kwargs)

        self.syslog_priorities = SYSLOG_IS_AVAILABLE and syslog_priorities

        self.colors_enabled = (
            terminal_supports_colors() and
            colors_are_enabled()
        )

    def format(self, record):
        current_thread_name = threading.current_thread().name

        # format record string
        time_stamp = datetime.datetime.fromtimestamp(record.created)
        time_stamp_str = time_stamp.strftime('%H:%M:%S.%f')

        record_string = '{}{} {}{} {} {} {}'.format(  # NOQA: FS002
            current_thread_name,
            (30 - len(current_thread_name)) * ' ',

            record.levelname,
            (8 - len(record.levelname)) * ' ',

            time_stamp_str,
            record.name,
            record.getMessage(),
        )

        # format exc_info
        if record.exc_info:
            record_string = '{}\n{}'.format(  # NOQA: FS002
                record_string,
                indent(
                    ''.join(format_exception(*record.exc_info))[:-1],
                    prefix='  ',
                ),
            )

        # syslog priorities
        if self.syslog_priorities:
            syslog_priority = get_syslog_priority(record.levelno)

            record_string = f'<{syslog_priority}>{record_string}'

        # colors
        if record.levelname == 'INFO' or not self.colors_enabled:
            return record_string

        RED = '31'
        YELLOW = '33'
        WHITE = '37'
        GREEN = '32'
        BACKGROUND_RED = '41'

        BRIGHT = '1'

        style = ''
        background = ''
        color = ''

        if record.levelname == 'DEBUG':
            style = ''
            background = ''
            color = GREEN

        elif record.levelname == 'WARNING':
            style = ''
            background = ''
            color = YELLOW

        elif record.levelname == 'ERROR':
            style = BRIGHT
            background = ''
            color = RED

        elif record.levelname == 'CRITICAL':
            style = BRIGHT
            background = BACKGROUND_RED
            color = WHITE

        if style:
            color = f';{color}'

        if background and color:
            background = f';{background}'

        return f'\033[{style}{color}{background}m{record_string}\033[00m'


def setup_logging(args):
    # set log level
    if not args.debug_mode:
        log_level = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warn': logging.WARN,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
        }[args.log_level.lower()]

    else:
        log_level = logging.DEBUG

    logging.basicConfig(level=log_level)

    # syslog priorities
    if args.syslog_priorities == 'always':
        syslog_priorities = True

    elif args.syslog_priorities == 'no':
        syslog_priorities = False

    elif args.syslog_priorities == 'auto':
        syslog_priorities = journald_is_running()

    # setup log formatting and log filtering
    log_formatter = LogFormatter(syslog_priorities=syslog_priorities)
    log_filter = LogFilter()

    for handler in logging.getLogger().root.handlers:
        handler.setFormatter(log_formatter)
        handler.addFilter(log_filter)

    if args.debug_mode:
        if args.debug_mode == 'messages':
            log_filter.include('lona.server.websockets')

        elif args.debug_mode == 'views':
            log_filter.include('lona.views')

        elif args.debug_mode == 'input-events':
            log_filter.include('lona.input_events')

        elif args.debug_mode == 'view-events':
            log_filter.include('lona.view_events')

    elif args.loggers:
        for logger_name in args.loggers:
            if logger_name.startswith('_'):
                log_filter.exclude(logger_name[1:])

            else:
                if logger_name.startswith('+'):
                    logger_name = logger_name[1:]

                log_filter.include(logger_name)

    return log_formatter, log_filter



from __future__ import annotations

from typing import Union, Dict, Any
from enum import Enum
import json

from lona._json import dumps

NoneType = type(None)


class PROTOCOL(Enum):
    MESSAGE_PREFIX = 'lona:'


class EXIT_CODE(Enum):
    SUCCESS = 0
    INVALID_MESSAGE = 1
    INVALID_METHOD = 2


class METHOD(Enum):
    # issuer: client
    VIEW = 101
    INPUT_EVENT = 102
    INPUT_EVENT_ACK = 103
    CLIENT_ERROR = 104
    PING = 105

    # issuer: server
    REDIRECT = 201
    HTTP_REDIRECT = 202
    DATA = 203
    VIEW_START = 204
    VIEW_STOP = 205
    PONG = 206


class INPUT_EVENT_TYPE(Enum):
    CLICK = 301
    CHANGE = 302
    CUSTOM = 303
    FOCUS = 304
    BLUR = 305


class NODE_TYPE(Enum):
    NODE = 401
    TEXT_NODE = 402
    WIDGET = 403


class DATA_TYPE(Enum):
    HTML = 501
    HTML_TREE = 502
    HTML_UPDATE = 503


class PATCH_TYPE(Enum):
    ID_LIST = 601
    CLASS_LIST = 602
    STYLE = 603
    ATTRIBUTES = 604
    NODES = 605
    WIDGET_DATA = 606


class OPERATION(Enum):
    SET = 701
    RESET = 702
    ADD = 703
    CLEAR = 704
    INSERT = 705
    REMOVE = 706


ENUMS = [
    PROTOCOL,
    EXIT_CODE,
    METHOD,
    INPUT_EVENT_TYPE,
    NODE_TYPE,
    DATA_TYPE,
    PATCH_TYPE,
    OPERATION,
]


def get_enum_values() -> Dict[str, Dict[str, Any]]:
    enums = {}

    for enum in ENUMS:
        enum_values = {}

        for enum_value in enum:
            enum_values[enum_value.name] = enum_value.value

        enums[enum.__name__] = enum_values

    return enums


def decode_message(raw_message: str) -> tuple[
    EXIT_CODE,
    None | int,
    None | int,
    None | int,
    Union[
        None,
        tuple[str, None | dict],
        tuple[int, int | str],
        tuple[str, None],
    ],
]:
    """
    returns: (exit_code, window_id, view_runtime_id, method, payload)

    """

    def _invalid_message() -> tuple[EXIT_CODE, None, None, None, None]:
        return EXIT_CODE.INVALID_MESSAGE, None, None, None, None

    if not raw_message.startswith(PROTOCOL.MESSAGE_PREFIX.value):
        return _invalid_message()

    message = raw_message[len(PROTOCOL.MESSAGE_PREFIX.value):]

    try:
        message = json.loads(message)

    except json.JSONDecodeError:
        return _invalid_message()

    if not isinstance(message, list):
        return _invalid_message()

    if len(message) != 4:
        return _invalid_message()

    window_id, view_runtime_id, method, payload = message

    try:
        method = METHOD(method)
        message[2] = method

    except ValueError:
        return _invalid_message()

    if method != METHOD.PING:
        # window_id
        if not isinstance(message[0], int):
            return _invalid_message()

        # view_runtime_id
        if not isinstance(message[1], (str, NoneType)):
            return _invalid_message()

    # view
    if method == METHOD.VIEW:
        if not isinstance(payload[0], str):
            return _invalid_message()

        if not isinstance(payload[1], (dict, NoneType)):
            return _invalid_message()

        return (EXIT_CODE.SUCCESS, *message)

    # input event
    if method == METHOD.INPUT_EVENT:

        # event id
        if not isinstance(payload[0], int):
            return _invalid_message()

        # event type
        if isinstance(payload[1], int):
            try:
                payload[1] = INPUT_EVENT_TYPE(payload[1])

            except ValueError:
                return _invalid_message()

        elif not isinstance(payload[1], str):
            return _invalid_message()

        return (EXIT_CODE.SUCCESS, *message)

    # client error
    if method == METHOD.CLIENT_ERROR:
        if not isinstance(payload[0], str):
            return _invalid_message()

        return (EXIT_CODE.SUCCESS, *message)

    # ping
    if method == METHOD.PING:
        if (window_id is not None or
                view_runtime_id is not None or
                payload is not None):

            return _invalid_message()

        return (EXIT_CODE.SUCCESS, None, None, METHOD.PING, None)

    return (EXIT_CODE.INVALID_METHOD, *message)


def encode_input_event_ack(window_id, view_runtime_id, input_event_id):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.INPUT_EVENT_ACK, input_event_id],
    )


def encode_redirect(window_id, view_runtime_id, target_url):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.REDIRECT, target_url],
    )


def encode_http_redirect(window_id, view_runtime_id, target_url):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.HTTP_REDIRECT, target_url],
    )


def encode_data(window_id, view_runtime_id, title, data):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.DATA, [title, data]],
    )


def encode_view_start(window_id, view_runtime_id):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.VIEW_START, None],
    )


def encode_view_stop(window_id, view_runtime_id):
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [window_id, view_runtime_id, METHOD.VIEW_STOP, None],
    )


def encode_pong():
    return PROTOCOL.MESSAGE_PREFIX.value + dumps(
        [None, None, METHOD.PONG, None],
    )



from __future__ import annotations

from collections.abc import AsyncIterator, Iterator, Callable
from typing import AsyncContextManager, Dict, cast, Any
from asyncio import AbstractEventLoop, Future, sleep
from contextlib import asynccontextmanager
from dataclasses import dataclass
import webbrowser as _webbrowser
import time
import os

from aiohttp.test_utils import TestClient
from aiohttp.web import Application
import pytest

from lona.shell.shell import embed_shell
from lona.worker_pool import WorkerPool
from lona.server import LonaServer
from lona import LonaApp


@dataclass
class LonaContext:
    client: TestClient
    app: None | LonaApp
    server: LonaServer
    event_loop: AbstractEventLoop

    def make_url(self, path: str = '') -> str:
        """
        Takes a path and returns a full URL to the running test server.

        This method is necessary because the Lona pytest plugin spins up a test
        server that runs on an unprivileged random port, so URLs are not stable
        between test runs.

        :path: (str) path to to view
        """

        return str(self.client.make_url(path))

    def debug_interactive(
            self,
            webbrowser: bool = True,
            sync: bool = False,
            locals: Dict[str, Any] | None = None,
    ) -> Future | None:  # pragma: no cover

        """
        This method pauses the current test by starting an rlpython shell and
        starts a webbrowser that points to the currently running test server.
        The test continues when the rlpython shell is exited.

        Async Example:

            async def test_my_feature(lona_app_context):
                context = await lona_app_context(setup_app)

                await context.debug_interactive(locals=locals())

        Sync Example:

            async def test_my_feature(lona_app_context):
                context = await lona_app_context(setup_app)

                def run_test():
                    context.debug_interactive(
                        sync=True,
                        locals=locals(),
                    )

                context.event_loop.run_in_executor(None, run_test)

        :webbrowser: (bool) start a webbrowser that points to the test server
        :sync:       (bool) run blocking in the current thread
        :locals:     (dict|None) variable overrides for the rlpython shell
        """

        def _debug_interactive():
            # start browser
            if webbrowser:
                os.environ['DISPLAY'] = ':0'
                _webbrowser.open(self.make_url())

            # embed shell
            _locals = locals or {}

            _locals = {
                'server': self.server,
                'lona_context': self,
                **_locals,
            }

            embed_shell(
                server=self.server,
                locals=_locals,
            )

        if sync:
            _debug_interactive()

            return None

        return self.event_loop.run_in_executor(None, _debug_interactive)


@pytest.fixture()
def lona_app_context(request, aiohttp_client, event_loop):
    async def setup_lona_app_context(
            setup_app: Callable[[LonaApp], None],
            project_root: str = '',
    ) -> LonaContext:
        # setup lona app
        lona_app = LonaApp(project_root or str(request.fspath))
        setup_app(lona_app)

        # setup client
        async def setup_aiohttp_app() -> Application:
            lona_app.setup_server(loop=event_loop)

            return cast(Application, lona_app.aiohttp_app)

        client = await aiohttp_client(await setup_aiohttp_app())

        return LonaContext(
            client=client,
            app=lona_app,
            server=cast(LonaServer, lona_app.server),
            event_loop=event_loop,
        )

    return setup_lona_app_context


@pytest.fixture()
def lona_project_context(request, aiohttp_client, event_loop):
    async def setup_lona_project_context(
        project_root: str = '',
        settings: None | list[str] = None,
        settings_pre_overrides: None | dict = None,
        settings_post_overrides: None | dict = None,
    ) -> LonaContext:

        # setup aiohttp app
        server = None

        async def setup_aiohttp_app() -> Application:
            nonlocal server

            # setup lona server
            server = LonaServer(
                project_root=project_root or request.fspath,
                settings_paths=settings or [],
                settings_pre_overrides=settings_pre_overrides or {},
                settings_post_overrides=settings_post_overrides or {},
            )

            server._loop = event_loop
            server._worker_pool = WorkerPool(settings=server.settings)

            return server._app

        client = await aiohttp_client(await setup_aiohttp_app())

        return LonaContext(
            client=client,
            app=None,
            server=cast(LonaServer, server),
            event_loop=event_loop,
        )

    return setup_lona_project_context


def eventually(
        timeout: float = 5,
        interval: float = 1,
) -> Iterator[AsyncContextManager]:

    """
    Wait for expected state in async test.

    The function is meant to be iterated. Each time it returns new attempt.
    Attempt is an async context manager to wrap assertions.
    It suppresses all exceptions until time is out.
    If no exception is raised iteration stops.

    Example::

        counter = 0
        for attempt in eventually():
            async with attempt:
                counter += 1
                assert counter > 3

    :param timeout: time in seconds during which it produces new attempts
    :param interval: seconds to sleep between attempts
    """

    end_time = time.time() + timeout
    success = False

    while not success:
        context_manager_was_used = False

        @asynccontextmanager
        async def attempt() -> AsyncIterator[None]:
            nonlocal context_manager_was_used, success
            context_manager_was_used = True

            try:
                yield  # execute assertions

            except Exception:
                if time.time() > end_time:
                    raise
                else:
                    await sleep(interval)

            else:
                success = True

        yield attempt()

        if not context_manager_was_used:
            raise SyntaxError('context manager must be used')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

from lona import VERSION_STRING

setup(
    include_package_data=True,
    name='lona',
    version=VERSION_STRING,
    author='Florian Scherf',
    url='https://github.com/lona-web-org/lona',
    author_email='mail@florianscherf.de',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'aiohttp>=3,<4',
        'jinja2',
        'rlpython',
        'typing-extensions',
    ],
    scripts=[
        'bin/lona',
    ],
    entry_points={
        'pytest11': [
            'lona = lona.pytest',
        ],
    },
)



from threading import RLock

from lona.html.abstract_node import AbstractNode
from lona.html.patches import PatchStack
from lona.protocol import DATA_TYPE


class Document:
    def __init__(self):
        self.title = ''
        self.html = None
        self._lock = RLock()
        self._patch_stack = PatchStack()

    @property
    def lock(self):
        return self._lock

    @property
    def is_dirty(self):
        return self._patch_stack.has_patches()

    def add_patch(self, *args, **kwargs):
        self._patch_stack.add_patch(*args, **kwargs)

    # html ####################################################################
    def get_node(self, node_id):
        node = None
        nodes = []

        with self.lock:
            if self.html.id == node_id:
                node = self.html

            else:
                for _node in self.html.iter_nodes():
                    if _node.id == node_id:
                        node = _node

                        break

            if node is None:
                return []

            while node is not None:
                nodes.append(node)
                node = node.parent

        return nodes

    def serialize(self):
        if not self.html:
            return self.apply(html='')

        # HTML string
        if isinstance(self.html, str):
            return self.title, DATA_TYPE.HTML, self.html

        # HTML tree
        return self.title, DATA_TYPE.HTML_TREE, self.html._serialize()

    def apply(self, title='', html=None):
        if title:
            self.title = title

        if html is None:
            return self.title, None, None

        # HTML update
        elif html is self.html:
            if not self._patch_stack.has_patches():
                return self.title, None, None

            patches = self._patch_stack.get_patches()
            self._patch_stack.clear()

            return self.title, DATA_TYPE.HTML_UPDATE, patches

        # HTML
        else:
            self._patch_stack.clear()

            if isinstance(self.html, AbstractNode):
                self.html._set_document(None)

            # node tree
            if isinstance(html, AbstractNode):
                self.html = html

                self.html._set_document(self)

                return self.title, DATA_TYPE.HTML_TREE, self.html._serialize()

            # HTML string
            self.html = str(html)

            return self.title, DATA_TYPE.HTML, html



import shutil
import os

from lona.logging import setup_logging
from lona.server import LonaServer


def collect_static(args):
    def _print(*print_args, **print_kwargs):
        if args.silent:
            return

        print(*print_args, **print_kwargs)

    def _rm(path):
        _print(f'rm -rf {path}')

        if args.dry_run:
            return

        if os.path.isdir(path):
            shutil.rmtree(path)

        else:
            os.unlink(path)

    def _cp(source, destination):
        source_is_dir = os.path.isdir(source)

        if source_is_dir:
            _print(f'cp -r {source} {destination}')

        else:
            _print(f'cp {source} {destination}')

        if args.dry_run:
            return

        if source_is_dir:
            shutil.copytree(source, destination)

        else:
            shutil.copy(source, destination)

    # setup logging
    setup_logging(args)

    # setup server
    server = LonaServer(
        project_root=args.project_root,
        settings_paths=args.settings,
        settings_pre_overrides=args.settings_pre_overrides,
        settings_post_overrides=args.settings_post_overrides,
    )

    # check destination
    if not os.path.exists(args.destination):
        exit(f"'{args.destination}' does not exist")

    if not os.path.isdir(args.destination):
        exit(f"'{args.destination}' is no directory")

    # clean
    if args.clean:
        for name in os.listdir(args.destination):
            path = os.path.join(args.destination, name)
            _rm(path)

    # copy files
    for path, url in server._static_file_loader.get_paths():
        destination = os.path.join(args.destination, url)

        _cp(path, destination)



from __future__ import annotations

from lona.unique_ids import generate_unique_id
from lona.static_files import StaticFile
from lona.html.selector import Selector
from lona.state import State


class DummyLock:
    def __enter__(self, *args, **kwargs):
        pass

    def __exit__(self, *args, **kwargs):
        pass


class DummyDocument:
    @property
    def lock(self):
        return DummyLock()

    def add_patch(self, *args, **kwargs):
        pass


class AbstractNode:
    STATIC_FILES: list[StaticFile] = []
    _subclasses: list[type] = []

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)

        cls._subclasses.append(cls)

    def __copy__(self, *args, **kwargs):
        raise RuntimeError('copy is not supported')

    def __deepcopy__(self, memo):
        raise RuntimeError('deepcopy is not supported')

    @staticmethod
    def get_all_node_classes():
        return [
            AbstractNode,
            *AbstractNode._subclasses,
        ]

    # id ######################################################################
    @property
    def id(self):
        if not hasattr(self, '_id'):
            self._id = generate_unique_id(name_space='nodes')

        return self._id

    # parent ##################################################################
    @property
    def parent(self):
        if not hasattr(self, '_parent'):
            self._parent = None

        return self._parent

    def _set_parent(self, parent):
        self._parent = parent

    # root ####################################################################
    @property
    def root(self):
        node = self

        while True:
            if node.parent is None:
                break

            node = node.parent

        return node

    # document ################################################################
    @property
    def document(self):
        _document = getattr(self.root, '_document', None)

        if not _document:
            return DummyDocument()

        return _document

    def _set_document(self, document):
        if self.parent:
            raise RuntimeError('node is no root node')

        self._document = document

    # locking #################################################################
    @property
    def lock(self):
        document = self.document

        if not document:
            return DummyLock()

        return document.lock

    # state ###################################################################
    @property
    def state(self):
        if not hasattr(self, '_state'):
            self._state = State(
                initial_data={},
                node=self,
            )

        return self._state

    # input events ############################################################
    def handle_change(self, input_event):
        return input_event

    def handle_click(self, input_event):
        return input_event

    def handle_focus(self, input_event):
        return input_event

    def handle_blur(self, input_event):
        return input_event

    def handle_input_event(self, input_event):
        if input_event.name == 'change':
            return self.handle_change(input_event)

        elif input_event.name == 'click':
            return self.handle_click(input_event)

        elif input_event.name == 'focus':
            return self.handle_focus(input_event)

        elif input_event.name == 'blur':
            return self.handle_blur(input_event)

        return input_event

    # queries #################################################################
    def iter_nodes(self, node=None):
        if node is None:
            node = self

        if hasattr(node, 'nodes'):
            for child in node.nodes:
                yield child

                for sub_child in self.iter_nodes(child):
                    yield sub_child

    def query_selector(self, raw_selector_string):
        selector = Selector(raw_selector_string)

        with self.lock:
            for node in self.iter_nodes():
                if selector.match(node):
                    return node

    def query_selector_all(self, raw_selector_string):
        selector = Selector(raw_selector_string)
        nodes = []

        with self.lock:
            for node in self.iter_nodes():
                if selector.match(node):
                    nodes.append(node)

            return nodes

    def closest(self, raw_selector_string):
        selector = Selector(raw_selector_string)

        with self.lock:
            node = self.parent

            while node is not None:
                if selector.match(node):
                    break

                node = node.parent

        return node



import os

from rlpython.utils.argument_parser import ReplArgumentParser


class LonaStaticFilesCommand:
    """
    List all loaded static files
    """

    NAME = 'lona_static_files'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(
            repl=self.repl,
            prog='lona_static_files',
            description=self.__doc__,
        )

        argument_parser.add_argument('static-dir', nargs='?')

        argument_parser.add_argument(
            '-l',
            '--list-directories',
            action='store_true',
        )

        argument_parser.add_argument(
            '-r',
            '--resolve',
        )

        args = argument_parser.parse_args(argv[1:])

        server = self.repl.globals['server']

        # resolve
        if args.resolve:
            abs_path = server._static_file_loader.resolve_path(args.resolve)

            if not abs_path:
                return 1

            self.repl.write(f'{abs_path}\n')

            return

        # list directories
        if args.list_directories:
            for static_dir in server._static_file_loader.static_dirs:
                self.repl.write(f'{static_dir}\n')

            return

        # list static files
        static_files = {}

        for static_dir in server._static_file_loader.static_dirs:
            for root, _dirs, files in os.walk(static_dir):
                for _file in files:
                    name = os.path.join(root, _file)

                    if name in static_files:
                        continue

                    static_files[name] = static_dir

        static_file_list = []

        for name, static_dir in static_files.items():
            static_file_list.append(os.path.join(static_dir, name))

        static_file_list.sort()

        for static_file in static_file_list:
            self.repl.write(f'{static_file}\n')
